
/**
 * \file rpmdb/pkgio.c
 */

#include "system.h"

#if defined(HAVE_MACHINE_TYPES_H)
# include <machine/types.h>
#endif

#include <netinet/in.h>

#define	_RPMIOB_INTERNAL
#include <rpmiotypes.h>
#include <rpmio_internal.h>
#include <rpmcb.h>
#include <rpmbc.h>		/* XXX beecrypt base64 */
#include <rpmmacro.h>
#define	_RPMHKP_INTERNAL
#include <rpmhkp.h>
#include <rpmku.h>

#define	_RPMTAG_INTERNAL
#include "header_internal.h"

#include <rpmdb.h>
#include <pkgio.h>

#define	_RPMTS_INTERNAL
#include "rpmts.h"

#include <rpmxar.h>

#include "signature.h"
#include "debug.h"

#ifdef __cplusplus
GENfree(struct rpmlead *)
GENfree(rpmuint8_t **)
GENfree(rpmuint32_t *)
#endif	/* __cplusplus */

int _pkgio_debug = 0;

/**
 */
ssize_t timedRead(FD_t fd, /*@out@*/ void * bufptr, size_t length);
#define	timedRead	(ufdio->read)

/*===============================================*/
/** \ingroup header
 * Write (with unload) header to file handle.
 * @param fd		file handle
 * @param h		header
 * @retval msg		*msg failure msg
 * @return		RPMRC_OK on success
 */
static
rpmRC rpmWriteHeader(FD_t fd, /*@null@*/ Header h, /*@null@*/ const char ** msg)
{
    const void * uh = NULL;
    size_t nb;
    size_t length;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */

if (_pkgio_debug)
fprintf(stderr, "--> rpmWriteHeader(%p, %p, %p)\n", fd, h, msg);

    if (h == NULL) {
	if (msg)
	    *msg = xstrdup(_("write of NULL header"));
	goto exit;
    }

    uh = headerUnload(h, &length);
    if (uh == NULL) {
	if (msg)
	    *msg = xstrdup(_("headerUnload failed"));
	goto exit;
    }

    {   unsigned char * hmagic = NULL;
	size_t nmagic = 0;

	(void) headerGetMagic(NULL, &hmagic, &nmagic);
	nb = Fwrite(hmagic, sizeof(hmagic[0]), nmagic, fd);
	if (nb != nmagic || Ferror(fd)) {
	    if (msg)
		*msg = (nb > 0
			? xstrdup(_("short write of header magic"))
			: xstrdup(Fstrerror(fd)) );
	    goto exit;
	}
    }

    nb = Fwrite(uh, sizeof(char), length, fd);
    if (nb != length || Ferror(fd)) {
	if (msg)
	    *msg = (nb > 0
		    ? xstrdup(_("short write of header"))
		    : xstrdup(Fstrerror(fd)) );
	goto exit;
    }
    rc = RPMRC_OK;

exit:
    uh = _free(uh);
    return rc;
}

/*===============================================*/

rpmop rpmtsOp(rpmts ts, rpmtsOpX opx)
{
    rpmop op = NULL;

    if (ts != NULL && (int)opx >= 0 && (int)opx < RPMTS_OP_MAX)
	op = ts->ops + opx;
    return op;
}

pgpDigParams rpmtsPubkey(const rpmts ts)
{
    return pgpGetPubkey(rpmtsDig(ts));
}

rpmdb rpmtsGetRdb(rpmts ts)
{
    rpmdb rdb = NULL;
    if (ts != NULL) {
	rdb = ts->rdb;
    }
    return rdb;
}

rpmRC rpmtsFindPubkey(rpmts ts, void * _dig)
{
    static const char * _dbpath = NULL;
    HE_t he = (HE_t) memset(alloca(sizeof(*he)), 0, sizeof(*he));
    pgpDig dig = (pgpDig) (_dig ? _dig : rpmtsDig(ts));
    pgpDigParams sigp = pgpGetSignature(dig);
    pgpDigParams pubp = pgpGetPubkey(dig);
    rpmRC res = RPMRC_NOKEY;
    const char * pubkeysource = NULL;
    rpmhkp hkp = NULL;
    rpmbf awol;
    rpmiob iob = NULL;
    int krcache = 1;	/* XXX assume pubkeys are cached in keyutils keyring. */
int validate = 0;
    int xx;

if (_rpmhkp_debug)
fprintf(stderr, "--> %s(%p,%p)\n", __FUNCTION__, ts, _dig);

assert(dig != NULL);
assert(sigp != NULL);
assert(pubp != NULL);
assert(rpmtsDig(ts) == dig);
    if (ts->hkp == NULL)
	ts->hkp = rpmhkpNew(NULL, 0);
    hkp = rpmhkpLink(ts->hkp);
    awol = rpmbfLink(hkp->awol);

#if 0
if (_rpmhkp_debug)
fprintf(stderr, "==> find sig id %08x %08x ts pubkey id %08x %08x\n",
pgpGrab(sigp->signid, 4), pgpGrab(sigp->signid+4, 4),
pgpGrab(hkp->signid, 4), pgpGrab(hkp->signid+4, 4));
#endif

    /* Lazy free of previous pubkey if pubkey does not match this signature. */
    if (memcmp(sigp->signid, hkp->signid, sizeof(hkp->signid))) {
#if 0
if (_rpmhkp_debug)
fprintf(stderr, "*** free pkt %p[%d] id %08x %08x\n", hkp->pkt, hkp->pktlen, pgpGrab(hkp->signid, 4), pgpGrab(hkp->signid+4, 4));
#endif
	hkp->pkt = _free(hkp->pkt);
	hkp->pktlen = 0;
	memset(hkp->signid, 0, sizeof(hkp->signid));
    }

    /* Has this pubkey failed a previous lookup? */
    if (hkp->pkt == NULL && awol != NULL
     && rpmbfChk(awol, sigp->signid, sizeof(sigp->signid)))
	goto leave;

    /* Try keyutils keyring lookup. */
    if (hkp->pkt == NULL) {
	iob = NULL;
	switch (rpmkuFindPubkey(sigp, &iob)) {
	case RPMRC_NOTFOUND:
	case RPMRC_FAIL:
	case RPMRC_NOTTRUSTED:
	case RPMRC_NOKEY:
	case RPMRC_NOSIG:
	    break;
	case RPMRC_OK:
	    krcache = 0;	/* XXX don't bother caching. */
	    hkp->pkt = (rpmuint8_t *)
		memcpy(xmalloc(iob->blen), iob->b, iob->blen);
	    hkp->pktlen = iob->blen;
	    pubkeysource = xstrdup("keyutils");
validate = 0;
	    break;
	}
if (_rpmhkp_debug)
fprintf(stderr, "\t%s: rpmku  %p[%u]\n", __FUNCTION__, hkp->pkt, (unsigned) hkp->pktlen);
    }

    /* Try rpmdb keyring lookup. */
    if (_dbpath == NULL)
	_dbpath = rpmGetPath("%{?_dbpath}", NULL);
    if (hkp->pkt == NULL && *_dbpath != '\0') {
	unsigned hx = 0xffffffff;
	unsigned ix = 0xffffffff;
	rpmmi mi;
	Header h;

	/* XXX Do a lazy open if not done already. */
	if (ts->rdb == NULL) {
	    xx = rpmdbOpen(ts->rootDir, &ts->rdb, ts->dbmode, (mode_t)0644);
	    if (xx)
		rpmlog(RPMLOG_ERR,
			_("cannot open Packages database in %s\n"), _dbpath);
	}

	/* Retrieve the pubkey that matches the signature. */
	he->tag = RPMTAG_PUBKEYS;
	mi = rpmmiInit(rpmtsGetRdb(ts), RPMTAG_PUBKEYS, sigp->signid, sizeof(sigp->signid));
	while ((h = rpmmiNext(mi)) != NULL) {
	    if (!headerGet(h, he, 0))
		continue;
	    hx = rpmmiInstance(mi);
	    switch (he->t) {
	    default:
		ix = 0xffffffff;
		break;
	    case RPM_STRING_ARRAY_TYPE:
		ix = he->c - 1;	/* XXX FIXME: assumes last pubkey */
		if (b64decode(he->p.argv[ix], (void **)&hkp->pkt, &hkp->pktlen))
		    ix = 0xffffffff;
		break;
	    }
	    he->p.ptr = _free(he->p.ptr);
	    break;
	}
	mi = rpmmiFree(mi);

	if (ix < 0xffffffff) {
	    char hnum[64];
	    sprintf(hnum, "h#%u[%u]", hx, ix);
	    pubkeysource = xstrdup(hnum);
validate = -1;	/* XXX rpmhkpValidate is prerequisite for rpmhkpFindKey */
	} else {
	    hkp->pkt = _free(hkp->pkt);
	    hkp->pktlen = 0;
	}
if (_rpmhkp_debug)
fprintf(stderr, "\t%s: rpmdb  %p[%u]\n", __FUNCTION__, hkp->pkt, (unsigned) hkp->pktlen);
    }

    /* Try autosign package pubkey (if present). */
    if (hkp->pkt == NULL && dig->pub && dig->publen > 0) {
	uint8_t keyid[8];

        xx = pgpPubkeyFingerprint((const rpmuint8_t *)dig->pub, dig->publen, keyid);
	if (!memcmp(sigp->signid, keyid, sizeof(keyid))) {
	    hkp->pkt = (uint8_t *) dig->pub;	dig->pub = NULL;
	    hkp->pktlen = dig->publen;		dig->publen = 0;
	    pubkeysource = xstrdup("package");
validate = -1;	/* XXX rpmhkpValidate is prerequisite for rpmhkpFindKey */
if (_rpmhkp_debug)
fprintf(stderr, "\t%s: auto   %p[%u]\n", __FUNCTION__, hkp->pkt, (unsigned) hkp->pktlen);
	}
    }

    /* Try keyserver lookup. */
    if (hkp->pkt == NULL) {
	const char * fn = rpmExpand("%{_hkp_keyserver_query}", "0x",
			pgpHexStr(sigp->signid, sizeof(sigp->signid)), NULL);

	xx = (fn && *fn != '%')
	    ? (pgpReadPkts(fn, &hkp->pkt, &hkp->pktlen) != PGPARMOR_PUBKEY)
	    : 1;	/* XXX assume failure */
	fn = _free(fn);
	if (xx) {
	    hkp->pkt = _free(hkp->pkt);
	    hkp->pktlen = 0;
	} else {
	    /* Save new pubkey in local ts keyring for delayed import. */
	    pubkeysource = xstrdup("keyserver");
validate = 1;
	}
if (_rpmhkp_debug)
fprintf(stderr, "\t%s: rpmhkp %p[%u]\n", __FUNCTION__, hkp->pkt, (unsigned) hkp->pktlen);
    }

#ifdef	NOTYET
    /* Try filename from macro lookup. */
    if (hkp->pkt == NULL) {
	const char * fn = rpmExpand("%{_gpg_pubkey}", NULL);

	xx = 0;
	if (fn && *fn != '%')
	    xx = (pgpReadPkts(fn, &hkp->pkt, &hkp->pktlen) != PGPARMOR_PUBKEY);
	fn = _free(fn);
	if (xx) {
	    hkp->pkt = _free(hkp->pkt);
	    hkp->pktlen = 0;
	} else {
	    pubkeysource = xstrdup("macro");
	}
    }
#endif

    /* Was a matching pubkey found? */
    if (hkp->pkt == NULL || hkp->pktlen == 0)
	goto exit;
if (_rpmhkp_debug)
fprintf(stderr, "\t%s: match  %p[%u]\n", __FUNCTION__, hkp->pkt, (unsigned) hkp->pktlen);

    /* Split the result into packet array. */
hkp->pkts = _free(hkp->pkts);	/* XXX memleaks */
hkp->npkts = 0;
    xx = pgpGrabPkts(hkp->pkt, hkp->pktlen, &hkp->pkts, &hkp->npkts);

    if (!xx)
        xx = pgpPubkeyFingerprint(hkp->pkt, hkp->pktlen, hkp->keyid);
    memcpy(pubp->signid, hkp->keyid, sizeof(pubp->signid)); /* XXX useless */

    /* Validate pubkey self-signatures. */
    /* XXX need at least 3 packets to validate a pubkey */
    if (validate) {
	rpmRC rc = rpmhkpValidate(hkp, NULL);
	switch (rc) {
	case RPMRC_OK:
	    break;
	case RPMRC_NOKEY:
	    if (validate < 0)	/* XXX ignore NOKEY w rpmdb pubkey. */
		break;
	    /*@fallthrough@*/
	case RPMRC_NOTFOUND:
	case RPMRC_FAIL:	/* XXX remap to NOTFOUND? */
	case RPMRC_NOTTRUSTED:
	default:
if (_rpmhkp_debug)
fprintf(stderr, "*** rpmhkpValidate: validate %d rc %d\n", validate, rc);
	    res = rc;
	    goto exit;
	}
    }

    /* Retrieve parameters from pubkey/subkey packet(s). */
    xx = rpmhkpFindKey(hkp, dig, sigp->signid, sigp->pubkey_algo);

#ifdef	DYING
if (_rpmhkp_debug)
_rpmhkpDumpDig(__FUNCTION__, dig, NULL);
#endif

    /* Do the parameters match the signature? */
    if (sigp->pubkey_algo == pubp->pubkey_algo
     &&	!memcmp(sigp->signid, pubp->signid, sizeof(sigp->signid)) )
    {

	/* Save the pubkey in the keyutils keyring. */
	if (krcache) {
	    if (iob == NULL) {
		iob = rpmiobNew(hkp->pktlen);
		iob->b = (rpmuint8_t *)memcpy(iob->b, hkp->pkt, iob->blen);
	    }
	    (void) rpmkuStorePubkey(sigp, iob);
if (_rpmhkp_debug)
fprintf(stderr, "\t%s: rpmku  %p[%u]\n", __FUNCTION__, hkp->pkt, (unsigned) hkp->pktlen);
	}

	/* Pubkey packet looks good, save the signer id. */
	memcpy(hkp->signid, pubp->signid, sizeof(hkp->signid));

	if (pubkeysource)
	    rpmlog(RPMLOG_DEBUG, "========== %s pubkey id %08x %08x (%s)\n",
		_pgpPubkeyAlgo2Name(sigp->pubkey_algo),
		pgpGrab(sigp->signid, 4), pgpGrab(sigp->signid+4, 4),
		pubkeysource);

	res = RPMRC_OK;

    }

exit:
    pubkeysource = _free(pubkeysource);
    if (res != RPMRC_OK) {
	hkp->pkt = _free(hkp->pkt);
	hkp->pktlen = 0;
	if (awol)
	    xx = rpmbfAdd(awol, sigp->signid, sizeof(sigp->signid));
    }

leave:
    (void) rpmbfFree(awol);
    (void) rpmhkpFree(hkp);

if (_rpmhkp_debug)
fprintf(stderr, "<-- %s(%p,%p) res %d\n", __FUNCTION__, ts, _dig, res);

    return res;
}

pgpDig rpmtsDig(rpmts ts)
{
    if (ts->dig == NULL) {
	ts->dig = pgpDigNew(RPMVSF_DEFAULT, (pgpPubkeyAlgo)0);
	(void) pgpSetFindPubkey(ts->dig, (int (*)(void *, void *))rpmtsFindPubkey, ts);
    }
    return ts->dig;
}

void rpmtsCleanDig(rpmts ts)
{
    if (ts && ts->dig) {
	rpmtsOpX opx;
	opx = RPMTS_OP_DIGEST;
	(void) rpmswAdd(rpmtsOp(ts, opx), (rpmop)pgpStatsAccumulator(ts->dig, opx));
	opx = RPMTS_OP_SIGNATURE;
	(void) rpmswAdd(rpmtsOp(ts, opx), (rpmop)pgpStatsAccumulator(ts->dig, opx));
	(void) pgpDigFree(ts->dig);
	ts->dig = NULL;		/* XXX make sure the ptr is __REALLY__ gone */
    }
}

/*===============================================*/

/**
 * The lead data structure.
 * The lead needs to be 8 byte aligned.
 * @deprecated The lead (except for signature_type) is legacy.
 * @todo Don't use any information from lead.
 */
struct rpmlead {
    unsigned char magic[4];
    unsigned char major;
    unsigned char minor;
    unsigned short type;
    unsigned short archnum;
    char name[66];
    unsigned short osnum;
    unsigned short signature_type; /*!< Signature type (RPMSIG_HEADERSIG) */
    char reserved[16];		/*!< Pad to 96 bytes -- 8 byte aligned! */
} ;

static unsigned char lead_magic[] = {
    0xed, 0xab, 0xee, 0xdb, 0x00, 0x00, 0x00, 0x00
};

/* The lead needs to be 8 byte aligned */

/**
 * Write lead to file handle.
 * @param fd		file handle
 * @param ptr		package lead
 * @param msg		*msg name to include in lead (or NULL)
 * @return		RPMRC_OK on success, RPMRC_FAIL on error
 */
static rpmRC wrLead(FD_t fd, const void * ptr, /*@null@*/ const char ** msg)
{
    struct rpmlead l = {};

if (_pkgio_debug)
fprintf(stderr, "--> wrLead(%p, %p, %p)\n", fd, ptr, msg);

    if (ptr)
	memcpy(&l, ptr, sizeof(l));

    /* Set some sane defaults */
    if ((int)l.major == 0)
	l.major = (unsigned char) 3;
    if (l.signature_type == 0)
	l.signature_type = 5;		/* RPMSIGTYPE_HEADERSIG */
    if (msg && *msg)
	(void) strncpy(l.name, *msg, sizeof(l.name));
    
    memcpy(&l.magic, lead_magic, sizeof(l.magic));
    l.type = (unsigned short) htons(l.type);
    l.archnum = (unsigned short) htons(l.archnum);
    l.osnum = (unsigned short) htons(l.osnum);
    l.signature_type = (unsigned short) htons(l.signature_type);
	
    if (Fwrite(&l, 1, sizeof(l), fd) != sizeof(l))
	return RPMRC_FAIL;

    return RPMRC_OK;
}

/**
 * Read lead from file handle.
 * @param fd		file handle
 * @retval ptr		*ptr package lead
 * @retval msg		*msg failure msg
 * @return		rpmRC return code
 */
static rpmRC rdLead(FD_t fd, /*@out@*/ /*@null@*/ void * ptr,
		/*@null@*/ const char ** msg)
{
static const char origin[] = "lead ";	/* XXX +trailing space */
    struct rpmlead ** leadp = (struct rpmlead **) ptr;
    struct rpmlead * l = (struct rpmlead *) xcalloc(1, sizeof(*l));
    char buf[BUFSIZ];
    static const rpmRC rc_fail = RPMRC_FAIL;
    static const rpmRC rc_notfound = RPMRC_NOTFOUND;
    rpmRC rc = rc_fail;		/* assume failure */
    int xx;

if (_pkgio_debug)
fprintf(stderr, "--> rdLead(%p, %p, %p)\n", fd, ptr, msg);

    buf[0] = '\0';
    if (leadp != NULL) *leadp = NULL;

    /* Read the first 96 bytes of the file. */
    if ((xx = (int) timedRead(fd, (char *)l, sizeof(*l))) != (int) sizeof(*l)) {
	if (Ferror(fd)) {
	    (void) snprintf(buf, sizeof(buf),
		_("%ssize(%u): BAD, read(%d), %s(%d)"),
		origin, (unsigned)sizeof(*l), xx, Fstrerror(fd), errno);
	    rc = rc_fail;
	} else {
	    (void) snprintf(buf, sizeof(buf),
		_("%ssize(%u): BAD, read(%d), %s(%d)"),
		origin, (unsigned)sizeof(*l), xx, strerror(errno), errno);
	    rc = rc_notfound;	/* XXX silently ignore a short read */
	}
	goto exit;
    }

    /* Attach a XAR handler on magic string "xar!" */
    {	rpmxar xar = fdGetXAR(fd);
	/* Attach rpmxar handler to fd if this is a xar archive. */
	if (xar == NULL) {
	    unsigned char * bh = (unsigned char *)l;
	    if (bh[0] == 'x' && bh[1] == 'a' && bh[2] == 'r' && bh[3] == '!') {
		const char * fn = fdGetOPath(fd);
assert(fn != NULL);
		xar = rpmxarNew(fn, "r");
		fdSetXAR(fd, xar);
		(void) rpmxarFree(xar, "rdLead");
	    }
	}

	/* With XAR, read from a xar archive file called "Lead". */
	xar = fdGetXAR(fd);
	if (xar != NULL) {
	    unsigned char *b = NULL;
	    size_t nb = 0;
	    const char item[] = "Lead";
	    if ((xx = rpmxarNext(xar)) != 0
	     || (xx = rpmxarPull(xar, item)) != 0)
	    {
		(void) snprintf(buf, sizeof(buf),
			_("XAR file not found (or no XAR support)"));
		rc = rc_notfound;	/* XXX silently ignore */
		goto exit;
	    }
	    (void) rpmxarSwapBuf(xar, NULL, 0, &b, &nb);
	    if (nb != sizeof(*l)) {
		(void) snprintf(buf, sizeof(buf),
			_("%ssize(%u): BAD, xar read(%u)"),
			origin, (unsigned)sizeof(*l), (unsigned)nb);
		b = _free(b);
		rc = rc_fail;
		goto exit;
	    }
	    memcpy(l, b, nb);
	    b = _free(b);
	}
    }

    l->type = (unsigned short) ntohs(l->type);
    l->archnum = (unsigned short) ntohs(l->archnum);
    l->osnum = (unsigned short) ntohs(l->osnum);
    l->signature_type = (unsigned short) ntohs(l->signature_type);

    if (memcmp(l->magic, lead_magic, sizeof(l->magic))) {
	(void) snprintf(buf, sizeof(buf), _("lead magic: BAD, read %02x%02x%02x%02x"), l->magic[0], l->magic[1], l->magic[2], l->magic[3]);
	rc = rc_notfound;	/* XXX silently ignore bad magic */
	goto exit;
    }

    switch (l->major) {
    default:
	(void) snprintf(buf, sizeof(buf),
		_("%sversion(%u): UNSUPPORTED"), origin, (unsigned) l->major);
	rc = rc_notfound;	/* XXX silently ignore invalid major */
	goto exit;
	/*@notreached@*/ break;
    case 3:
    case 4:
	break;
    }

    if (l->signature_type != 5) {	/* RPMSIGTYPE_HEADERSIG */
	(void) snprintf(buf, sizeof(buf),
		_("%stype(%u): UNSUPPORTED"), "sigh ", (unsigned) l->signature_type);
	rc = rc_notfound;	/* XXX silently ignore signature type */
	goto exit;
    }

    rc = RPMRC_OK;

exit:
    if (rc == RPMRC_OK && leadp != NULL)
	*leadp = l;
    else
	l = _free(l);
	
    if (msg != NULL && buf[0] != '\0') {
	buf[sizeof(buf)-1] = '\0';
	*msg = xstrdup(buf);
    }
    return rc;
}

/*===============================================*/

/**
 * Write signature header.
 * @param fd		file handle
 * @param ptr		signature header
 * @retval msg		*msg failure msg
 * @return		rpmRC return code
 */
static rpmRC wrSignature(FD_t fd, void * ptr,
		/*@unused@*/ /*@null@*/ const char ** msg)
{
    Header sigh = (Header) ptr;
    static unsigned char zero[8]
	= { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' };
    size_t sigSize;
    size_t pad;
    rpmRC rc = RPMRC_OK;

if (_pkgio_debug)
fprintf(stderr, "--> wrSignature(%p, %p, %p)\n", fd, ptr, msg);

    rc = rpmWriteHeader(fd, sigh, msg);
    if (rc != RPMRC_OK)
	return rc;

    sigSize = headerSizeof(sigh);
    pad = (8 - (sigSize % 8)) % 8;
    if (pad) {
	if (Fwrite(zero, sizeof(zero[0]), pad, fd) != pad)
	    rc = RPMRC_FAIL;
    }
    rpmlog(RPMLOG_DEBUG, D_("Signature: size(%u)+pad(%u)\n"), (unsigned)sigSize, (unsigned)pad);
    return rc;
}

/**
 * Print package size.
 * @param fd			package file handle
 * @param siglen		signature header size
 * @param pad			signature padding
 * @param datalen		length of header+payload
 * @retval *st			stat(2) of input file
 * @return 			rpmRC return code
 */
static inline rpmRC printSize(FD_t fd, size_t siglen, size_t pad,
		size_t datalen, struct stat *st)
{
    size_t expected;
    size_t nl = rpmpkgSizeof("Lead", NULL);
    int lvl = RPMLOG_DEBUG;
    rpmRC rc = RPMRC_OK;	/* assume success */

    if (st == NULL)
	st = memset(alloca(sizeof(*st)), 0, sizeof(*st));

#ifndef	DYING	/* XXX Fstat(2) contentLength not gud enuf yet. */
    int fdno = Fileno(fd);
    /* HACK: workaround for davRead wiring. */
    if (fdno == 123456789) {
	st->st_size = 0;
	st->st_size -= nl + siglen + pad + datalen;
    } else
#endif
    if (Fstat(fd, st) < 0)
	return RPMRC_FAIL;

    expected = nl + siglen + pad + datalen;
    if (fdno != 123456789) {
	if (expected == (size_t)st->st_size)
	    goto exit;
	/* XXX rpm-3.0.x built packages retrofit an immutable region */
	expected -= 2 * REGION_TAG_COUNT;
	if (expected == (size_t)st->st_size)
	    goto exit;
	lvl = RPMLOG_ERR;
	rc = RPMRC_FAIL;
    }
exit:
    rpmlog(lvl,
	D_("Expected size: %12lu = lead(%u)+sigs(%u)+pad(%u)+data(%lu)\n"),
		(unsigned long)expected,
		(unsigned)nl, (unsigned) siglen, (unsigned) pad,
		(unsigned long)datalen);
    rpmlog(lvl,
	D_("  Actual size: %12lu\n"), (unsigned long)st->st_size);

    return rc;
}

static char * entryInfoStr(entryInfo info)
{
    static char buf[128];
    (void) snprintf(buf, sizeof(buf),
	_("tag %u type %u offset %d count %u"),
		(unsigned) info->tag,
		(unsigned) info->type,
		(int) info->offset,
		(unsigned) info->count);
    return buf;
}

static rpmRC chkPgpPktLen(const char * origin, rpmuint32_t tag,
		const void *sig, rpmuint32_t siglen, pgpPkt pp,
		char *b, size_t nb)
{
    int xx = pgpPktLen(sig, siglen, pp);
    rpmRC rc = RPMRC_OK;

    if (xx < 0 || xx != (int)siglen) {
	(void) snprintf(b, nb,
		_("%stag(%u): BAD, OpenPGP count(%u) != pktlen(%d)"),
		origin, tag, siglen, xx);
	rc = RPMRC_FAIL;
    }

    return rc;
}

/**
 * Read (and verify header+payload size) signature header.
 * @param fd		file handle
 * @retval ptr		*ptr signature header (or NULL)
 * @retval msg		*msg failure msg
 * @return		rpmRC return code
 */
static rpmRC rdSignature(FD_t fd, /*@out@*/ /*@null@*/ void * ptr,
		/*@null@*/ const char ** msg)
{
static const char origin[] = "sigh ";	/* XXX +trailing space */
    HE_t he = (HE_t) memset(alloca(sizeof(*he)), 0, sizeof(*he));
    Header * sighp = (Header *) ptr;
    char buf[BUFSIZ];
    rpmuint32_t block[4];
    rpmuint32_t il;
    rpmuint32_t dl;
    rpmuint32_t * ei = NULL;
    entryInfo pe;
    size_t pvlen;
    size_t startoff;
    size_t nb;
    rpmuint32_t ril = 0;
    indexEntry entry = (indexEntry)
	memset(alloca(sizeof(*entry)), 0, sizeof(*entry));
    entryInfo info = (entryInfo)
	memset(alloca(sizeof(*info)), 0, sizeof(*info));
    unsigned char * dataStart;
    unsigned char * dataEnd = NULL;
    Header sigh = NULL;
    rpmRC rc = RPMRC_FAIL;		/* assume failure */
    int xx;
    struct stat sb, *st = &sb;
    static int map = 1;
pgpDig dig = pgpDigLink(fdGetDig(fd));
rpmuint32_t sigtag = 0;
rpmuint32_t sigtype = 0;
const void * sig = NULL;
rpmuint32_t siglen = 0;
pgpPkt pp = (pgpPkt) alloca(sizeof(*pp));
rpmuint32_t rtag = 0;	/* assume no region tag. */

if (_pkgio_debug)
fprintf(stderr, "--> rdSignature(%p, %p, %p) dig %p\n", fd, ptr, msg, dig);

    buf[0] = '\0';
    if (sighp)
	*sighp = NULL;

    /* With XAR, read from a xar archive file called "Signature". */
    {	rpmxar xar = fdGetXAR(fd);
	if (xar != NULL) {
	    const char item[] = "Signature";
	    if ((xx = rpmxarNext(xar)) != 0
	     || (xx = rpmxarPull(xar, item)) != 0)
	    {
		(void) snprintf(buf, sizeof(buf),
			_("XAR file not found (or no XAR support)"));
		rc = RPMRC_NOTFOUND;
		goto exit;
	    }
	}
    }

    if (Fstat(fd, st) < 0) {
	(void) snprintf(buf, sizeof(buf),
		_("%sstat: BAD, Fstat(2) failed"), origin);
	goto exit;
    }

    memset(block, 0, sizeof(block));
    startoff = fd->stats->ops[FDSTAT_READ].bytes;
    if ((xx = (int) timedRead(fd, (char *)block, sizeof(block))) != (int) sizeof(block)) {
	(void) snprintf(buf, sizeof(buf),
		_("%ssize(%d): BAD, read returned %d"),
		origin, (int)sizeof(block), xx);
	goto exit;
    }

    {   unsigned char * hmagic = NULL;
	size_t nmagic = 0;

	(void) headerGetMagic(NULL, &hmagic, &nmagic);

	if (memcmp(block, hmagic, nmagic)) {
	    unsigned char * x = (unsigned char *)block;
	    (void) snprintf(buf, sizeof(buf),
		_("%smagic: BAD, read %02x%02x%02x%02x%02x%02x%02x%02x"),
		origin, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7]);
	    goto exit;
	}
    }
    /* XXX arbitrary max limit check doesn't help much */
    /* XXX rpm-3.0.x built packages have only 2 tags */
    il = (rpmuint32_t) ntohl(block[2]);
    if (il > (st->st_size - startoff - sizeof(block)) || il < 2 || il > 16) {
	(void) snprintf(buf, sizeof(buf),
		_("%stags: BAD, no. of tags(%u) out of range"),
		origin, (unsigned) il);
	goto exit;
    }
    /* XXX arbitrary max limit check doesn't help much */
    /* XXX rpm-3.0.x built packages have only 20 bytes */
    dl = (rpmuint32_t) ntohl(block[3]);
    if (dl > (st->st_size - startoff - sizeof(block)) || dl < (4+16) || dl > 8192) {
	(void) snprintf(buf, sizeof(buf),
		_("%sdata: BAD, no. of bytes(%u) out of range"),
		origin, (unsigned) dl);
	goto exit;
    }

    nb = (il * sizeof(struct entryInfo_s)) + dl;
    if (nb > (st->st_size - startoff - sizeof(block))) {
	(void) snprintf(buf, sizeof(buf),
		_("%sblob: BAD, header size (%u) larger than file size"),
		origin, (unsigned) nb);
       goto exit;
    }

    pvlen = (sizeof(il) + sizeof(dl) + nb);
    if (map) {
        static const int prot = PROT_READ | PROT_WRITE;
        static const int flags = MAP_PRIVATE| MAP_ANONYMOUS;
        static const int fdno = -1;
        static const off_t off = 0;

	ei = (rpmuint32_t *) mmap(NULL, pvlen, prot, flags, fdno, off);
assert(ei != NULL && ei != (void *)-1);		/* coverity #1214082 */
	if (ei == NULL || ei == (void *)-1)
            fprintf(stderr,
                "==> mmap(%p[%u], 0x%x, 0x%x, %d, 0x%x) error(%d): %s\n",
                NULL, (unsigned)pvlen, prot, flags, fdno, (unsigned)off,
                errno, strerror(errno));
    } else
    {
	ei = (rpmuint32_t *) xmalloc(pvlen);
    }

    if ((xx = (int) timedRead(fd, (char *)&ei[2], nb)) != (int) nb) {
	(void) snprintf(buf, sizeof(buf),
		_("%sblob(%u): BAD, read returned %d"), origin, (unsigned) nb, xx);
	goto exit;
    }
    ei[0] = block[2];
    ei[1] = block[3];

    if (map) {
        if (mprotect(ei, pvlen, PROT_READ) != 0)
            fprintf(stderr, "==> mprotect(%p[%u],0x%x) error(%d): %s\n",
                        ei, (unsigned)pvlen, PROT_READ,
                        errno, strerror(errno));
    }

    pe = (entryInfo) &ei[2];
    dataStart = (unsigned char *) (pe + il);
    
    /* Check (and convert) the 1st tag element. */
    xx = headerVerifyInfo(1, dl, pe, &entry->info, 0);
    if (xx != -1 ||
     !(  (entry->info.tag == RPMTAG_HEADERSIGNATURES
       && entry->info.type == REGION_TAG_TYPE
       && entry->info.offset > 0
       && entry->info.count == (rpmTagCount)REGION_TAG_COUNT)
      || (entry->info.tag == (rpmTag)RPMSIGTAG_SIZE
       && entry->info.type == RPM_UINT32_TYPE
       && entry->info.offset == 0
       && entry->info.count == (rpmTagCount)1)
      ))
    {
	(void) snprintf(buf, sizeof(buf),
		_("tag[%d]: BAD, %s"), 0, entryInfoStr(&entry->info));
	goto exit;
    }

    /* Is there an immutable header region tag? */
    if (entry->info.tag == RPMTAG_HEADERSIGNATURES)
    {
	rtag = RPMTAG_HEADERSIGNATURES;
	/* Is the region tag sane? */
	if (!(entry->info.type == REGION_TAG_TYPE
	   && entry->info.count == (rpmTagCount)REGION_TAG_COUNT))
	{
	    (void) snprintf(buf, sizeof(buf),
		_("region tag: BAD, %s"), entryInfoStr(&entry->info));
	    goto exit;
	}

/*
 * XXX http://mysql.mirrors.pair.com/Downloads/MySQL-5.0/MySQL-client-community-5.0.51a-0.rhel4.i386.rpm
 * built by rpm-4.3.3 (from REL4) has entry->info.offset == 0.
 */
assert(entry->info.offset >= 0);	/* XXX insurance */
	if (entry->info.offset >= (rpmint32_t)dl) {
	    (void) snprintf(buf, sizeof(buf),
		_("region offset: BAD, %s"), entryInfoStr(&entry->info));
	    goto exit;
	}

	/* Is there an immutable header region tag trailer? */
	dataEnd = dataStart + entry->info.offset;
	(void) memcpy(info, dataEnd, REGION_TAG_COUNT);
	/* XXX Really old packages have HEADER_IMAGE, not HEADER_SIGNATURES. */
	if (info->tag == (rpmuint32_t) htonl(RPMTAG_HEADERIMAGE)) {
	    rpmuint32_t stag = (rpmuint32_t) htonl(RPMTAG_HEADERSIGNATURES);
	    info->tag = (rpmTag) stag;
#ifdef	DYING	/* XXX can't retrofit with PROT_READ */
	    memcpy(dataEnd, &stag, sizeof(stag));
#endif
	}
	dataEnd += REGION_TAG_COUNT;

	xx = headerVerifyInfo(1, il * sizeof(*pe), info, &entry->info, 1);
	if (xx != -1 ||
	    !(entry->info.tag == RPMTAG_HEADERSIGNATURES
	   && entry->info.type == REGION_TAG_TYPE
	   && entry->info.offset > 0
	   && entry->info.count == (rpmTagCount)REGION_TAG_COUNT))
	{
	    (void) snprintf(buf, sizeof(buf),
		_("region trailer: BAD, %s"), entryInfoStr(&entry->info));
	    goto exit;
	}
	memset(info, 0, sizeof(*info));

	/* Is the no. of tags in the region less than the total no. of tags? */
	ril = (rpmuint32_t) (entry->info.offset/sizeof(*pe));
	if ((entry->info.offset % sizeof(*pe)) || ril > il) {
	    (void) snprintf(buf, sizeof(buf),
		_("region size: BAD, ril(%u) > il(%u)"), (unsigned) ril, (unsigned) il);
	    goto exit;
	}
    }

    /* Sanity check signature tags */
    memset(info, 0, sizeof(*info));
    xx = headerVerifyInfo(il-1, dl, pe+1, &entry->info, 0);
    if (xx != -1) {
	(void) snprintf(buf, sizeof(buf),
		_("%stag[%d]: BAD, %s"), origin, xx, entryInfoStr(&entry->info));
	goto exit;
    }

    /* OK, blob looks sane, load the header. */
    sigh = headerLoad(ei);
    if (sigh == NULL) {
	(void) snprintf(buf, sizeof(buf), _("%sload: BAD"), origin);
	goto exit;
    }
    if (map) {
	sigh->flags |= HEADERFLAG_MAPPED;
	sigh->flags |= HEADERFLAG_RDONLY;
    } else
	sigh->flags |= HEADERFLAG_ALLOCATED;
    sigh->flags |= HEADERFLAG_SIGNATURE;

    {	size_t sigSize = headerSizeof(sigh);
	size_t pad = (8 - (sigSize % 8)) % 8; /* 8-byte pad */

	/* Position at beginning of header. */
	if (pad && (xx = (int) timedRead(fd, (char *)block, pad)) != (int) pad)
	{
	    (void) snprintf(buf, sizeof(buf),
		_("%spad(%u): BAD, read %d bytes"), origin, (unsigned) pad, xx);
	    goto exit;
	}

	/* If sigh had a region, then sigh should have a RPMSIGTAG_SHA1. */
	if (rtag) {
	    he->tag = (rpmTag) RPMSIGTAG_SHA1;
	    xx = headerGet(sigh, he, HEADERGET_SIGHEADER);
	    he->p.ptr = _free(he->p.ptr);
	    if (!xx) {
		(void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, missing %s tag"), origin,
			he->tag, "SHA1");
		rc = RPMRC_FAIL;
		goto exit;
	    }
	}

	/* All packages should have RPMSIGTAG_MD5. */
	he->tag = (rpmTag) RPMSIGTAG_MD5;
	xx = headerGet(sigh, he, HEADERGET_SIGHEADER);
	he->p.ptr = _free(he->p.ptr);
	if (!xx) {
	    (void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, missing %s tag"), origin,
			he->tag, "MD5");
	    rc = RPMRC_FAIL;
	    goto exit;
	}

	/* All packages should have RPMSIGTAG_SIZE. */
	he->tag = (rpmTag) RPMSIGTAG_SIZE;
	xx = headerGet(sigh, he, HEADERGET_SIGHEADER);
	if (xx && he->t == RPM_UINT32_TYPE && he->c == 1) {
	    size_t datasize = he->p.ui32p[0];
	    he->p.ptr = _free(he->p.ptr);
	    /* Print package component sizes. */
	    rc = printSize(fd, sigSize, pad, datasize, st);
	    if (rc != RPMRC_OK) {
		(void) snprintf(buf, sizeof(buf),
			_("%ssize(%u): BAD, file length check failed"),
			origin, (unsigned) sigSize);
		goto exit;
	    }
	} else {
	    he->p.ptr = _free(he->p.ptr);
	    (void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, missing %s tag"), origin,
			he->tag, "SIZE");
	    rc = RPMRC_FAIL;
	    goto exit;
	}

    }
    (void) headerSetStartOff(sigh, (rpmuint32_t)startoff);
    (void) headerSetEndOff(sigh, fd->stats->ops[FDSTAT_READ].bytes);

    HeaderIterator hi;
    rc = RPMRC_OK;	/* assume success */

    for (hi = headerInit(sigh);
        headerNext(hi, he, HEADERGET_SIGHEADER);
        he->p.ptr = _free(he->p.ptr))
    {
	rpmTagType tagtype;
	rpmTagCount tagcount;
	rpmuint32_t tagsize;
	unsigned checkit;
	unsigned savesig;

	tagtype = 0;
	tagcount = 0;
	tagsize = 0;
	checkit = 0;
	savesig = 0;

assert(he->p.ptr != NULL);
	switch ((rpmSigTag)he->tag) {
	case RPMSIGTAG_GPG:
	case RPMSIGTAG_PGP5:
	case RPMSIGTAG_PGP:
	    tagtype = RPM_BIN_TYPE;
	    checkit = 1;
	    break;
	case RPMSIGTAG_SHA1:
	    tagtype = RPM_STRING_TYPE;
	    tagcount = 1;
	    tagsize = 2 * 160/8 + 1;
	    checkit = 2;
	    if (sig == NULL)	/* XXX prefer signature over digest */
		savesig = 1;	/* XXX saved iff dig != NULL */
	    break;
	case RPMSIGTAG_MD5:
	    tagtype = RPM_BIN_TYPE;
	    tagcount = 128/8;
//	    tagsize = 128/8;
	    break;
	case RPMSIGTAG_RSA:
	case RPMSIGTAG_DSA:
	case RPMSIGTAG_ECDSA:
	    tagtype = RPM_BIN_TYPE;
	    checkit = 1;
#ifdef	NOTYET
	    savesig = 1;	/* XXX saved iff dig != NULL */
#endif
	    break;
	case RPMSIGTAG_SIZE:
	case RPMSIGTAG_PAYLOADSIZE:
	    tagtype = RPM_UINT32_TYPE;
	    tagcount = 1;
//	    tagsize = tagcount * sizeof(rpmuint32_t);
	    break;
#ifdef	REFERENCE
	case RPMSIGTAG_LONGSIZE:
	case RPMSIGTAG_LONGARCHIVESIZE:
	    tagtype = RPM_UINT64_TYPE;
	    tagcount = 1;
//	    tagsize = tagcount * sizeof(rpmuint64_t);
	    break;
	case RPMSIGTAG_RESERVED:
	    tagtype = RPM_BIN_TYPE;
	    break;
#endif
	case RPMSIGTAG_PADDING:
	    tagtype = RPM_BIN_TYPE;
	    checkit = 3;
	    break;
	default:
	    continue;
	    break;
	}


#ifdef	NOTYET
	/* XXX needs entry->length: augment HE_t? */
	if (tagsize && entry->length != tagsize) {
	    (void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, invalid size(%u)"),
			origin, he->tag, entry->length);
	    rc = RPMRC_FAIL;
	    break;
	}
#endif

	if (tagcount && he->c != tagcount) {
	    (void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, invalid count(%u)"),
			origin, he->tag, he->c);
	    rc = RPMRC_FAIL;
	    break;
	}

	/* XXX header_internal.c handles type verification. */
	if (tagtype && he->t != tagtype) {
	    (void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, invalid type(%u)"),
			origin, he->tag, he->t);
	    rc = RPMRC_FAIL;
	    break;
	}

	switch (checkit) {
	case 1:		/* Verify he->c == OpenPGP pgpPktLen. */
	    /* XXX todo++: check the entire packet? */
	    xx = chkPgpPktLen(origin, he->tag, he->p.ui8p, he->c, pp,
			buf, sizeof(buf));
	    if (xx != RPMRC_OK) {
		rc = RPMRC_FAIL;
		break;
	    }
	    break;
	case 2:		/* Verify hex string. */
	{   const char * s = he->p.str;
	    rpmuint32_t ns = 0;
	    while (ns < tagsize && *s != '\0') {
		if (strchr("0123456789abcdefABCDEF", *s) == NULL)
		    break;
		ns++;
		s++;
	    }
	    if (ns != (tagsize-1) || *s != '\0') {
		s = he->p.str;
		(void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, not hex len(%u) \"%*s\""),
			origin, he->tag, ns, ns, s);
		rc = RPMRC_FAIL;
		break;
	    }
	}   break;
	case 3:		/* Verify all zero's */
	{   const rpmuint8_t * b = he->p.ui8p;
	    rpmuint32_t nb;
	    for (nb = 0; nb < he->c; nb++) {
		if (b[nb] != 0)
		    break;
	    }
	    if (nb < he->c) {
		(void) snprintf(buf, sizeof(buf),
			_("%stag(%u): BAD, not zero at byte offset(%u)"),
			origin, he->tag, nb);
		rc = RPMRC_FAIL;
		break;
	    }
	}   break;
	default:
	    break;
	}
	if (rc != RPMRC_OK)
	    break;

	/* XXX saved iff dig != NULL (i.e. fdSetDig(fd) has been done) */
	if (dig && savesig) {
	    sigtag = he->tag;
	    sigtype = he->t;
	    sig = _free(sig);
	    sig = he->p.ptr;
	    he->p.ptr = NULL;	/* XXX steal the malloc'd data */
	    siglen = he->c;
	}

    }
    he->p.ptr = _free(he->p.ptr);
    hi = headerFini(hi);

exit:

    if (rc == RPMRC_OK && dig && sig && siglen > 0) {
	(void) pgpSetSig(dig, sigtag, sigtype, sig, siglen);
	sig = NULL;	/* XXX avoid free */
    }
    sig = _free(sig);
    if (sighp)
	*sighp = ((sigh && rc == RPMRC_OK) ?  headerLink(sigh) : NULL);
    (void)headerFree(sigh);
    sigh = NULL;

    if (sig)
	sig = _free(sig);
    dig = pgpDigFree(dig);

    if (msg != NULL) {
	buf[sizeof(buf)-1] = '\0';
	*msg = xstrdup(buf);
    }

if (_pkgio_debug)
fprintf(stderr, "<-- rdSignature(%p, %p, %p) rc %d \"%s\"\n", fd, ptr, msg, rc, *msg);

    return rc;
}

/*===============================================*/
static int hBlobDigest(const void * uh, pgpDig dig, pgpHashAlgo hash_algo,
	const unsigned char * regionEnd, rpmuint32_t ril, DIGEST_CTX * ctxp)
{
    rpmuint32_t * ei = (rpmuint32_t *) uh;
    rpmuint32_t il = (rpmuint32_t) ntohl(ei[0]);
    rpmuint32_t dl = (rpmuint32_t) ntohl(ei[1]);
    entryInfo pe = (entryInfo) &ei[2];
    rpmuint32_t ildl[2];
    const unsigned char * dataStart = (const unsigned char *) (pe + il);
    rpmop op;
    unsigned char * b = NULL;
    size_t nb = 0;
    int xx = 0;

    (void)dl;

    ildl[0] = (rpmuint32_t) htonl(ril);
    ildl[1] = (rpmuint32_t) (regionEnd - dataStart);
    ildl[1] = (rpmuint32_t) htonl(ildl[1]);

    op = (rpmop) pgpStatsAccumulator(dig, 10);	/* RPMTS_OP_DIGEST */
    (void) rpmswEnter(op, 0);
    *ctxp = rpmDigestInit(hash_algo, RPMDIGEST_NONE);

    (void) headerGetMagic(NULL, &b, &nb);
    if (b && nb > 0) {
	(void) rpmDigestUpdate(*ctxp, b, nb);
	dig->nbytes += nb;
    }

    b = (unsigned char *) ildl;
    nb = sizeof(ildl);
    (void) rpmDigestUpdate(*ctxp, b, nb);
    dig->nbytes += nb;

    b = (unsigned char *) pe;
    nb = (size_t) (htonl(ildl[0]) * sizeof(*pe));
    (void) rpmDigestUpdate(*ctxp, b, nb);
    dig->nbytes += nb;

    b = (unsigned char *) dataStart;
    nb = (size_t) htonl(ildl[1]);
    (void) rpmDigestUpdate(*ctxp, b, nb);
    dig->nbytes += nb;
    (void) rpmswExit(op, dig->nbytes);

    return xx;
}

static rpmRC hBlobVerify(const void * uh, pgpDig dig,
	indexEntry entry,
	const unsigned char * regionEnd, rpmuint32_t ril,
	char *b, size_t nb)
{
    rpmuint32_t * ei = (rpmuint32_t *) uh;
#ifdef	UNUSED
    rpmuint32_t il = (rpmuint32_t) ntohl(ei[0]);
#endif
    rpmuint32_t dl = (rpmuint32_t) ntohl(ei[1]);
    entryInfo pe = (entryInfo) &ei[2];
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    int xx;
entryInfo info = &entry->info;
const void * sig = pgpGetSig(dig);
size_t siglen = dig->siglen;
pgpPkt pp = (pgpPkt) alloca(sizeof(*pp));
DIGEST_CTX * ctxp = NULL;

    /* Do SHA1 if possible). */
    switch (dig->sigtag) {
    case RPMTAG_RSAHEADER:
    case RPMTAG_DSAHEADER:
    case RPMTAG_ECDSAHEADER:
	/* Parse the parameters from the OpenPGP packets that will be needed. */
	xx = chkPgpPktLen("", dig->sigtag, sig, siglen, pp, b, nb);
	if (xx != RPMRC_OK) {
	    rc = RPMRC_FAIL;
	    goto exit;
	}
	xx = rpmhkpLoadSignature(NULL, dig, pp);
	if (xx < 0) {
	    (void) snprintf(b, nb,
		_("skipping header with unverifiable V%u signature"),
		(unsigned) dig->signature.version);
	    rc = RPMRC_FAIL;
	    goto exit;
	}

	switch (dig->signature.pubkey_algo) {
	default:
	    (void) snprintf(b, nb,
		_("skipping header with unknown signature algorithm(%u)\n"),
		dig->signature.pubkey_algo);
	    rc = RPMRC_FAIL;
	    goto exit;
	    break;
	case PGPPUBKEYALGO_RSA:		ctxp = &dig->hrsa;	break;
	case PGPPUBKEYALGO_DSA:		ctxp = &dig->hdsa;	break;
	case PGPPUBKEYALGO_ECDSA:	ctxp = &dig->hecdsa;	break;
	}

	xx = hBlobDigest(uh, dig, dig->signature.hash_algo,
			regionEnd, ril, ctxp);

	break;
    case RPMSIGTAG_SHA1:
	ctxp = &dig->hdsa;	/* XXX dig->hsha? */
	xx = hBlobDigest(uh, dig, PGPHASHALGO_SHA1,
			regionEnd, ril, ctxp);
	break;
    default:
	xx = (ril > 0 ? headerVerifyInfo(ril-1, dl, pe+1, info, 0) : -1);
	if (xx != -1) {
	    (void) snprintf(b, nb,
			_("tag[%d]: BAD, %s"),
			xx, entryInfoStr(info));
	    rc = RPMRC_FAIL;
	} else {
	    (void) snprintf(b, nb, "Header sanity check: OK");
	    rc = RPMRC_OK;
	}
	b[nb-1] = '\0';
	goto exit;
	break;
    }

    b[0] = '\0';
    rc = rpmVerifySignature(dig, b);

exit:
    if (ctxp && *ctxp) {
	(void) rpmDigestFinal(*ctxp, NULL, NULL, 0);
	*ctxp = NULL;
    }
    dig->sig = _free(dig->sig);

    return rc;
}

/**
 * Check header consistency, performing headerGet() the hard way.
 *
 * Sanity checks on the header are performed while looking for a
 * header-only digest or signature to verify the blob. If found,
 * the digest or signature is verified.
 *
 * @param dig		signature parameters container
 * @param uh		unloaded header blob
 * @param uc		no. of bytes in blob (or 0 to disable)
 * @retval msg		*msg signature verification msg
 * @return		RPMRC_OK/RPMRC_NOTFOUND/RPMRC_FAIL
 */
rpmRC headerCheck(pgpDig dig, const void * uh, size_t uc, const char ** msg)
{
static const char origin[] = "hdr ";	/* XXX +trailing space */
    char buf[8*BUFSIZ];
    rpmuint32_t * ei = (rpmuint32_t *) uh;
    rpmuint32_t il = (rpmuint32_t) ntohl(ei[0]);
    rpmuint32_t dl = (rpmuint32_t) ntohl(ei[1]);
    entryInfo pe = (entryInfo) &ei[2];
    rpmuint32_t ildl[2];
    size_t pvlen = sizeof(ildl) + (il * sizeof(*pe)) + dl;
    const unsigned char * dataStart = (const unsigned char *) (pe + il);
    indexEntry entry = (indexEntry)
	memset(alloca(sizeof(*entry)), 0, sizeof(*entry));
    entryInfo info = (entryInfo)
	memset(alloca(sizeof(*info)), 0, sizeof(*info));
    const void * sig = NULL;
    rpmVSFlags vsflags = pgpDigVSFlags;
    size_t siglen = 0;
    rpmuint32_t ril = 0;
    const unsigned char * regionEnd = NULL;
    rpmRC rc = RPMRC_FAIL;	/* assume failure */
    int xx;
    rpmuint32_t i;
pgpPkt pp = (pgpPkt) alloca(sizeof(*pp));

if (_pkgio_debug)
fprintf(stderr, "--> headerCheck(%p, %p[%u], %p)\n", dig, uh, (unsigned) uc, msg);

    buf[0] = '\0';

    /* Is the blob the right size? */
    if (uc > 0 && pvlen != uc) {
	(void) snprintf(buf, sizeof(buf),
		_("blob size(%d): BAD, 8 + 16 * il(%u) + dl(%u)"),
		(int)uc, (unsigned)il, (unsigned)dl);
	goto exit;
    }

    /* Check (and convert) the 1st tag element. */
    xx = headerVerifyInfo(1, dl, pe, &entry->info, 0);
    if (xx != -1) {
	(void) snprintf(buf, sizeof(buf),
		_("tag[%d]: BAD, %s"), 0, entryInfoStr(&entry->info));
	goto exit;
    }

    /* Is there an immutable header region tag? */
    if (entry->info.tag != RPMTAG_HEADERIMMUTABLE) {
	if (entry->info.tag < RPMTAG_HEADERI18NTABLE) {
	    (void) snprintf(buf, sizeof(buf),
		_("region tag: BAD, %s"), entryInfoStr(&entry->info));
	    goto exit;
	}
	rc = RPMRC_NOTFOUND;
    }

    /* Is the region tag sane? */
    if (!(entry->info.type == RPM_BIN_TYPE
       && entry->info.count == (rpmTagCount)REGION_TAG_COUNT))
    {
	(void) snprintf(buf, sizeof(buf),
		_("region tag: BAD, %s"), entryInfoStr(&entry->info));
	goto exit;
    }

    /* Is the offset within the data area? */
    if (entry->info.offset >= (int) dl) {
	(void) snprintf(buf, sizeof(buf),
		_("region offset: BAD, %s"), entryInfoStr(&entry->info));
	goto exit;
    }

    /* Is there an immutable header region tag trailer? */
    regionEnd = dataStart + entry->info.offset;
    (void) memcpy(info, regionEnd, REGION_TAG_COUNT);
    regionEnd += REGION_TAG_COUNT;

    xx = headerVerifyInfo(1, il * sizeof(*pe), info, &entry->info, 1);
    if (xx != -1 ||
	!(entry->info.tag == RPMTAG_HEADERIMMUTABLE
       && entry->info.type == REGION_TAG_TYPE
       && entry->info.offset > 0
       && entry->info.count == (rpmTagCount)REGION_TAG_COUNT))
    {
	(void) snprintf(buf, sizeof(buf),
		_("region trailer: BAD, %s"), entryInfoStr(&entry->info));
	goto exit;
    }
    memset(info, 0, sizeof(*info));

    /* Is the no. of tags in the region less than the total no. of tags? */
    ril = (rpmuint32_t) (entry->info.offset/sizeof(*pe));
    if ((entry->info.offset % sizeof(*pe)) || ril > il) {
	(void) snprintf(buf, sizeof(buf),
		_("region size: BAD, ril(%u) > il(%u)"), (unsigned) ril, (unsigned)il);
	goto exit;
    }

    /* Fish out the autosign pubkey (if present). */
    memset(info, 0, sizeof(*info));
    for (i = 1; i < (unsigned) il; i++) {
        xx = headerVerifyInfo(1, dl, pe+i, &entry->info, 0);
	if (xx != -1)
	    continue;
	if (entry->info.tag == RPMTAG_PUBKEYS) {
 	    dig->pub = _free(dig->pub);
 	    dig->publen = 0;
 	    {   rpmiob iob = rpmiobNew(0);
		/* XXX FIXME: assumes first pubkey */
		const unsigned char *s = dataStart + entry->info.offset;
 		iob = rpmiobAppend(iob, (char *)s, 0);
 		xx = pgpArmorUnwrap(iob,
			(rpmuint8_t **)&dig->pub, &dig->publen);
 		iob = rpmiobFree(iob);
 	    }
 	    if (xx != PGPARMOR_PUBKEY) {
 		dig->pub = _free(dig->pub);
 		dig->publen = 0;
 	    }
 	    break;
	}
    }

    /* Find an appended header-only digest/signature tag. */
    for (i = ril; i < (unsigned) il; i++) {
	static const rpmuint32_t hsigmin = 16;
	static const rpmuint32_t hsigmax = 2048;
	xx = headerVerifyInfo(1, dl, pe+i, &entry->info, 0);
	if (xx != -1) {
	    (void) snprintf(buf, sizeof(buf),
		_("tag[%d]: BAD, %s"), i, entryInfoStr(&entry->info));
	    goto exit;
	}

	switch (entry->info.tag) {
	case RPMTAG_SHA1HEADER:
	{   const unsigned char * s;
	    int ns;
	    if (vsflags & RPMVSF_NOSHA1HEADER)
		break;
	    if (entry->info.type != RPM_STRING_TYPE || entry->info.count != 1)
	    {
		(void) snprintf(buf, sizeof(buf),
			_("%sSHA1: BAD, type(%u) != STRING count(%u) != 1"),
			origin, entry->info.type, entry->info.count);
		goto exit;
	    }
	    ns = 0;
	    for (s = dataStart + entry->info.offset; *s != '\0'; s++) {
		if (strchr("0123456789abcdefABCDEF", *s) == NULL)
		    break;
		ns++;
	    }
	    if (*s != '\0' || ns != 40) {
		s = dataStart + entry->info.offset;
		(void) snprintf(buf, sizeof(buf),
			_("%sSHA1: BAD, not hex off(%d) len(%u) \"%*s\""),
			origin, entry->info.offset, ns, ns, s);
		goto exit;
	    }
	    if (info->tag == 0) {
		*info = entry->info;	/* structure assignment */
		siglen = ns + 1;
	    }
	}   break;
	case RPMTAG_RSAHEADER:
	    if (vsflags & RPMVSF_NORSAHEADER)
		break;
	    if (entry->info.type != RPM_BIN_TYPE
	     || (entry->info.count < hsigmin || entry->info.count > hsigmax))
	    {
		(void) snprintf(buf, sizeof(buf),
			_("%sRSA: BAD, type(%u) != BINARY count(%u) != [%u,%u]"),
			origin, entry->info.type, entry->info.count, hsigmin, hsigmax);
		goto exit;
	    }
	    sig = dataStart + entry->info.offset;
	    xx = chkPgpPktLen(origin, entry->info.tag,
			sig, entry->info.count, pp, buf, sizeof(buf));
	    if (xx != RPMRC_OK) {
		rc = RPMRC_FAIL;
		goto exit;
	    }
	    *info = entry->info;	/* structure assignment */
	    siglen = info->count;
	    break;
	case RPMTAG_DSAHEADER:
	    if (vsflags & RPMVSF_NODSAHEADER)
		break;
	    if (entry->info.type != RPM_BIN_TYPE
	     || (entry->info.count < hsigmin || entry->info.count > hsigmax))
	    {
		(void) snprintf(buf, sizeof(buf),
			_("%sDSA: BAD, type(%u) != BINARY count(%u) != [%u,%u]"),
			origin, entry->info.type, entry->info.count, hsigmin, hsigmax);
		goto exit;
	    }
	    sig = dataStart + entry->info.offset;
	    xx = chkPgpPktLen(origin, entry->info.tag,
			sig, entry->info.count, pp, buf, sizeof(buf));
	    if (xx != RPMRC_OK) {
		rc = RPMRC_FAIL;
		goto exit;
	    }
	    *info = entry->info;	/* structure assignment */
	    siglen = info->count;
	    break;
	case RPMTAG_ECDSAHEADER:
	    if (vsflags & RPMVSF_NOECDSAHEADER)
		break;
	    if (entry->info.type != RPM_BIN_TYPE
	     || (entry->info.count < hsigmin || entry->info.count > hsigmax))
	    {
		(void) snprintf(buf, sizeof(buf),
			_("%sECDSA: BAD, type(%u) != BINARY count(%u) != [%u,%u]"),
			origin, entry->info.type, entry->info.count, hsigmin, hsigmax);
		goto exit;
	    }
	    sig = dataStart + entry->info.offset;
	    xx = chkPgpPktLen(origin, entry->info.tag,
			sig, entry->info.count, pp, buf, sizeof(buf));
	    if (xx != RPMRC_OK) {
		rc = RPMRC_FAIL;
		goto exit;
	    }
	    *info = entry->info;	/* structure assignment */
	    siglen = info->count;
	    break;
	default:
	    break;
	}
    }
    rc = RPMRC_NOTFOUND;

exit:

    switch (rc) {
    case RPMRC_OK:
    case RPMRC_FAIL:
    default:
	break;
    case RPMRC_NOTFOUND:
	/* Verify header-only digest/signature (or sanity check). */
	if (ril == 0)
	    ril = il;
	if (regionEnd == NULL)
	    regionEnd = ((const unsigned char *)uh) + uc;

assert(dig != NULL);
	dig->nbytes = 0;

	/* Add signature info for an appended tag. */
	if (siglen && dig->sigtag == 0) {
	    sig = memcpy(xmalloc(siglen), dataStart + info->offset, siglen);
	    (void) pgpSetSig(dig, info->tag, info->type, sig, siglen);
	}

	rc = hBlobVerify(uh, dig, entry, regionEnd, ril, buf, sizeof(buf));
	break;
    }

    buf[sizeof(buf)-1] = '\0';
    if (msg) *msg = xstrdup(buf);

if (_pkgio_debug)
fprintf(stderr, "<-- headerCheck#3: rc %d \"%s\"\n", rc, (msg ? *msg: ""));
    return rc;
}

/**
 * Return size of Header.
 * @param ptr		metadata header (at least 32 bytes)
 * @return		size of header
 */
RPM_GNUC_PURE
static size_t szHeader(/*@null@*/ const void * ptr)
{
    rpmuint32_t p[4];
assert(ptr != NULL);
    memcpy(p, ptr, sizeof(p));
    return (8 + 8 + 16 * ntohl(p[2]) + ntohl(p[3]));
}

/**
 * Check metadata header.
 * @param fd		file handle
 * @param ptr		metadata header
 * @retval msg		*msg failure msg
 * @return		rpmRC return code
 */
static rpmRC ckHeader(/*@unused@*/ FD_t fd, const void * ptr,
		/*@unused@*/ /*@null@*/ const char ** msg)
{
    rpmRC rc = RPMRC_OK;
    Header h;

    h = headerLoad((void *)ptr);
    if (h == NULL)
	rc = RPMRC_FAIL;
    (void)headerFree(h);
    h = NULL;

    return rc;
}

/** 
 * Return checked and loaded header.
 * @param fd		file handle
 * @retval hdrp		address of header (or NULL)
 * @retval msg		*msg verification error message (or NULL)
 * @return		RPMRC_OK on success
 */
static rpmRC rpmReadHeader(FD_t fd, /*@null@*/ Header * hdrp,
		/*@null@*/ const char ** msg)
{
    pgpDig dig = pgpDigLink(fdGetDig(fd));
    char buf[BUFSIZ];
    rpmuint32_t block[4];
    rpmuint32_t il;
    rpmuint32_t dl;
    rpmuint32_t * ei = NULL;
    size_t uc = 0;
    unsigned char * b;
    size_t startoff;
    size_t nb;
    Header h = NULL;
    const char * origin = NULL;
    rpmRC rc = RPMRC_FAIL;		/* assume failure */
    int xx;
    static int map = 1;
    struct stat sb, *st = &sb;

if (_pkgio_debug)
fprintf(stderr, "--> rpmReadHeader(%p, %p, %p)\n", fd, hdrp, msg);

    /* Create (if not already) a signature parameters container. */
    if (dig == NULL) {
	dig = pgpDigNew(RPMVSF_DEFAULT, (pgpPubkeyAlgo)0);
	(void) fdSetDig(fd, dig);
    }

    buf[0] = '\0';

    if (hdrp)
	*hdrp = NULL;

    /* With XAR, read from a xar archive file called "Header". */
    {	rpmxar xar = fdGetXAR(fd);
	if (xar != NULL) {
	    const char item[] = "Header";
	    if ((xx = rpmxarNext(xar)) != 0
	     || (xx = rpmxarPull(xar, item)) != 0)
	    {
		(void) snprintf(buf, sizeof(buf),
			_("XAR file not found (or no XAR support)"));
		rc = RPMRC_NOTFOUND;
		goto exit;
	    }
	}
    }

    if (Fstat(fd, st) < 0) {
	(void) snprintf(buf, sizeof(buf),
		_("hdr stat: BAD, Fstat(2) failed"));
	goto exit;
    }

    memset(block, 0, sizeof(block));
    startoff = fd->stats->ops[FDSTAT_READ].bytes;
    if ((xx = (int) timedRead(fd, (char *)block, sizeof(block))) != (int)sizeof(block)) {
	/* XXX Handle EOF's as RPMRC_NOTFOUND, not RPMRC_FAIL, returns. */
	if (xx == 0)
	    rc = RPMRC_NOTFOUND;
	else
	    (void) snprintf(buf, sizeof(buf),
		_("hdr size(%u): BAD, read returned %d"),
		(unsigned)sizeof(block), xx);
	goto exit;
    }

    b = NULL;
    nb = 0;
    (void) headerGetMagic(NULL, &b, &nb);
    if (memcmp(block, b, nb)) {
	unsigned char * x = (unsigned char *) block;
	(void) snprintf(buf, sizeof(buf),
		_("hdr magic: BAD, read %02x%02x%02x%02x%02x%02x%02x%02x"),
		x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7]);
	goto exit;
    }

    il = (rpmuint32_t)ntohl(block[2]);
    if (il < 1
     || (il * sizeof(struct entryInfo_s)) > (st->st_size - startoff - sizeof(block))
     || hdrchkTags(il))
    {
	(void) snprintf(buf, sizeof(buf),
		_("hdr tags: BAD, no. of tags(%u) out of range"),
		(unsigned) il);

	goto exit;
    }
    dl = (rpmuint32_t)ntohl(block[3]);
    if (dl < 1
     || dl > (st->st_size - startoff - sizeof(block))
     || hdrchkData(dl))
    {
	(void) snprintf(buf, sizeof(buf),
		_("hdr data: BAD, no. of bytes(%u) out of range"),
		(unsigned) dl);
	goto exit;
    }

    nb = (il * sizeof(struct entryInfo_s)) + dl;
    if (nb > (st->st_size - startoff - sizeof(block))) {
	(void) snprintf(buf, sizeof(buf),
		_("hdr blob: BAD, header size (%u) larger than file size"),
		(unsigned) nb);
       goto exit;
    }
    uc = sizeof(il) + sizeof(dl) + nb;
    if (map) {
        static const int prot = PROT_READ | PROT_WRITE;
        static const int flags = MAP_PRIVATE| MAP_ANONYMOUS;
        static const int fdno = -1;
        static const off_t off = 0;

	ei = (rpmuint32_t *) mmap(NULL, uc, prot, flags, fdno, off);
assert(ei != NULL && ei != (void *)-1);		/* coverity #124081 */
	if (ei == NULL || ei == (void *)-1)
            fprintf(stderr,
                "==> mmap(%p[%u], 0x%x, 0x%x, %d, 0x%x) error(%d): %s\n",
                NULL, (unsigned)uc, prot, flags, fdno, (unsigned)off,
                errno, strerror(errno));
    } else {
	ei = (rpmuint32_t *) xmalloc(uc);
    }

    if ((xx = (int) timedRead(fd, (char *)&ei[2], nb)) != (int) nb) {
	(void) snprintf(buf, sizeof(buf),
		_("hdr blob(%u): BAD, read returned %d"), (unsigned)nb, xx);
	goto exit;
    }
    ei[0] = block[2];
    ei[1] = block[3];

    if (map) {
        if (mprotect(ei, uc, PROT_READ) != 0)
            fprintf(stderr, "==> mprotect(%p[%u],0x%x) error(%d): %s\n",
                        ei, (unsigned)uc, PROT_READ,
                        errno, strerror(errno));
    }

    /* Sanity check header tags */
    rc = headerCheck(dig, ei, uc, msg);
    if (rc != RPMRC_OK)
	goto exit;

    /* OK, blob looks sane, load the header. */
    h = headerLoad(ei);
    if (h == NULL) {
	(void) snprintf(buf, sizeof(buf), _("hdr load: BAD"));
        goto exit;
    }
    if (map) {
	h->flags |= HEADERFLAG_MAPPED;
	h->flags |= HEADERFLAG_RDONLY;
    } else
	h->flags |= HEADERFLAG_ALLOCATED;
    ei = NULL;	/* XXX will be freed with header */

    /* Save the opened path as the header origin. */
    /* XXX TODO: push the Realpath() underneath fdGetOPath(). */
    origin = fdGetOPath(fd);
    if (origin != NULL) {
	const char * lpath = NULL;
	int ut = urlPath(origin, &lpath);
	(void)ut;	/* XXX keep gcc/clang quiet. */
	if (lpath && *lpath != '/') {
	    char * rpath = Realpath(origin, NULL);
	    (void) headerSetOrigin(h, (rpath ? rpath : NULL));
	    rpath = _free(rpath);
	} else
	    (void) headerSetOrigin(h, origin);
    }
    /* Save file size in metadata header. */
    {	struct stat * st = headerGetStatbuf(h);
	int saveno = errno;
	(void) Fstat(fd, st);
	errno = saveno;
    }
    (void) headerSetStartOff(h, (rpmuint32_t)startoff);
    (void) headerSetEndOff(h, fd->stats->ops[FDSTAT_READ].bytes);
    
exit:
    if (hdrp && h && rc == RPMRC_OK)
	*hdrp = headerLink(h);
    if (ei != NULL && uc > 0) {
	if (map) {
            if (munmap(ei, uc) != 0)
                fprintf(stderr, "==> munmap(%p[%u]) error(%d): %s\n",
                ei, (unsigned)uc, errno, strerror(errno));
	    ei = NULL;
	} else
	    ei = _free(ei);
    }
    dig = pgpDigFree(dig);
    (void)headerFree(h);
    h = NULL;

    if (msg != NULL && *msg == NULL && buf[0] != '\0') {
	buf[sizeof(buf)-1] = '\0';
	*msg = xstrdup(buf);
    }

if (_pkgio_debug)
fprintf(stderr, "<-- rpmReadHeader: rc %d \"%s\"\n", rc, (msg ? *msg: ""));
    return rc;
}

/**
 * Read metadata header.
 * @param fd		file handle
 * @retval ptr		*ptr metadata header (or NULL)
 * @retval msg		*msg failure msg
 * @return		rpmRC return code
 */
static rpmRC rdHeader(FD_t fd, /*@out@*/ /*@null@*/ void * ptr,
		/*@null@*/ const char ** msg)
{
    Header * hdrp = (Header *) ptr;
    return rpmReadHeader(fd, hdrp, msg);
}

/**
 * Write metadata header.
 * @param fd		file handle
 * @param ptr		metadata header
 * @retval msg		*msg failure msg
 * @return		rpmRC return code
 */
static rpmRC wrHeader(FD_t fd, void * ptr,
		/*@unused@*/ /*@null@*/ const char ** msg)
{
    Header h = (Header) ptr;
    return rpmWriteHeader(fd, h, msg);
}

/*===============================================*/

size_t rpmpkgSizeof(const char * fn, const void * ptr)
{
    size_t len = 0;

    if (!strcmp(fn, "Lead"))
	len = 96;	/* RPMLEAD_SIZE */
    else
    if (!strcmp(fn, "Signature")) {
	len = szHeader(ptr);
	len += ((8 - (len % 8)) % 8);   /* padding */
    } else
    if (!strcmp(fn, "Header"))
	len = szHeader(ptr);
    return len;
}

rpmRC rpmpkgCheck(const char * fn, FD_t fd, const void * ptr, const char ** msg)
{
    rpmRC rc = RPMRC_FAIL;

    if (msg)
	*msg = NULL;

    if (!strcmp(fn, "Header"))
	rc = ckHeader(fd, ptr, msg);
    return rc;
}

rpmRC rpmpkgRead(const char * fn, FD_t fd, void * ptr, const char ** msg)
{
    rpmRC rc = RPMRC_FAIL;

    if (msg)
	*msg = NULL;

    if (!strcmp(fn, "Lead"))
	rc = rdLead(fd, ptr, msg);
    else
    if (!strcmp(fn, "Signature"))
	rc = rdSignature(fd, ptr, msg);
    else
    if (!strcmp(fn, "Header"))
	rc = rdHeader(fd, ptr, msg);
    return rc;
}

rpmRC rpmpkgWrite(const char * fn, FD_t fd, void * ptr, const char ** msg)
{
    rpmRC rc = RPMRC_FAIL;

    if (msg)
	*msg = NULL;

    if (!strcmp(fn, "Lead"))
	rc = wrLead(fd, ptr, msg);
    else
    if (!strcmp(fn, "Signature"))
	rc = wrSignature(fd, ptr, msg);
    else
    if (!strcmp(fn, "Header"))
	rc = wrHeader(fd, ptr, msg);
    return rc;
}
